use std::sync::Arc;

use geoffrey_db::helper::load_location;
use geoffrey_db::query::QueryBuilder;
use geoffrey_models::models::item::ItemListing;
use geoffrey_models::models::locations::{Location, LocationDataDb, LocationDb, LocationType};
use geoffrey_models::models::parameters::add_item_params::AddItemParams;
use geoffrey_models::models::player::Player;
use geoffrey_models::models::response::api_error::GeoffreyAPIError;
use geoffrey_models::models::settings::GeoffreySettings;
use geoffrey_models::models::CommandLevel;

use crate::api_endpoint::{ApiEndpoint, RequestType};
use crate::commands::Command;
use crate::context::Context;
use crate::helper::validate_string_parameter;
use crate::Result;

pub struct AddItem {}

impl ApiEndpoint for AddItem {
    fn endpoint_name() -> String {
        "add_item".to_string()
    }

    fn request_type() -> RequestType {
        RequestType::POST
    }
}

impl Command for AddItem {
    type Req = AddItemParams;
    type Resp = Location;

    fn command_level() -> CommandLevel {
        CommandLevel::REGISTERED
    }

    fn run_command(ctx: Arc<Context>, req: &Self::Req, user: Option<Player>) -> Result<Self::Resp> {
        let user = user.unwrap();

        let query = QueryBuilder::<LocationDb>::default()
            .with_type(LocationType::Shop)
            .with_owner(user.id.unwrap())
            .with_name(&req.shop);

        let mut shop = ctx
            .db
            .run_query(query)?
            .pop()
            .ok_or(GeoffreyAPIError::EntryNotFound)?;

        if let LocationDataDb::Shop(shop_data) = &mut shop.loc_data {
            shop_data.item_listings.insert(ItemListing::new(
                &req.item_name,
                req.price,
                req.quantity,
            ));

            let shop = ctx.db.insert(shop)?;

            load_location(&ctx.db, &shop).map_err(|err| err.into())
        } else {
            Err(GeoffreyAPIError::EntryNotFound)
        }
    }

    fn validate_parameters(req: &Self::Req, cfg: &GeoffreySettings) -> Result<()> {
        if req.quantity == 0 {
            return Err(GeoffreyAPIError::ParameterInvalid("quantity".to_string()));
        }

        validate_string_parameter("item_name", &req.item_name, cfg.max_str_len)?;

        Ok(())
    }
}
