use async_trait::async_trait;
use reqwest::Method;
use serenity::builder::CreateApplicationCommand;
use serenity::model::interactions::application_command::{
    ApplicationCommandInteraction, ApplicationCommandOptionType,
};

use geoffrey_models::models::locations::Location;
use geoffrey_models::models::parameters::delete_params::DeleteParams;
use geoffrey_models::models::response::api_error::GeoffreyAPIError;

use crate::bot::arg_parse::option_to_string;
use crate::bot::commands::BotCommand;
use crate::bot::lang::PLAYER_DOES_NOT_HAVE_MATCHING_LOC;
use crate::context::GeoffreyContext;
use crate::error::BotError;
use serenity::utils::MessageBuilder;

pub struct DeleteCommand;

#[async_trait]
impl BotCommand for DeleteCommand {
    type ApiParams = DeleteParams;
    type ApiResp = Location;

    fn command_name() -> String {
        "delete".to_string()
    }

    fn request_type() -> Method {
        Method::POST
    }

    fn custom_err_resp(err: &BotError) -> Option<String> {
        match err {
            BotError::GeoffreyApi(GeoffreyAPIError::EntryNotFound) => {
                Some(PLAYER_DOES_NOT_HAVE_MATCHING_LOC.to_string())
            }
            _ => None,
        }
    }

    fn create_app_command(command: &mut CreateApplicationCommand) -> &mut CreateApplicationCommand {
        command
            .name(Self::command_name())
            .description("Delete a location from Geoffrey")
            .create_option(|option| {
                option
                    .name("loc_name")
                    .description("Name of the location")
                    .kind(ApplicationCommandOptionType::String)
                    .required(true)
            })
    }

    async fn process_arguments(
        command_interaction: ApplicationCommandInteraction,
    ) -> Result<Self::ApiParams, BotError> {
        let options = command_interaction.data.options;
        let loc_name = option_to_string(options.get(0), "loc_name")?;

        Ok(Self::ApiParams::new(loc_name))
    }

    fn build_response(_: &GeoffreyContext, resp: Self::ApiResp, _: Self::ApiParams) -> String {
        MessageBuilder::new()
            .push_bold_safe(resp.name)
            .push(" has been removed from Geoffrey, good riddance!")
            .build()
    }
}
