use async_trait::async_trait;
use reqwest::Method;
use serenity::builder::CreateApplicationCommand;
use serenity::model::interactions::application_command::{
    ApplicationCommandInteraction, ApplicationCommandOptionType,
};

use geoffrey_models::models::locations::Location;
use geoffrey_models::models::parameters::item_command_params::ItemCommandParams;
use geoffrey_models::models::response::api_error::GeoffreyAPIError;

use crate::bot::arg_parse::option_to_string;
use crate::bot::commands::BotCommand;
use crate::bot::lang::PLAYER_DOES_NOT_HAVE_MATCHING_SHOP;
use crate::context::GeoffreyContext;
use crate::error::BotError;
use serenity::utils::MessageBuilder;

pub struct RemoveItemCommand;

#[async_trait]
impl BotCommand for RemoveItemCommand {
    type ApiParams = ItemCommandParams;
    type ApiResp = Location;

    fn command_name() -> String {
        "remove_item".to_string()
    }

    fn request_type() -> Method {
        Method::POST
    }

    fn custom_err_resp(e: &BotError) -> Option<String> {
        match e {
            BotError::GeoffreyApi(GeoffreyAPIError::EntryNotFound) => {
                Some(PLAYER_DOES_NOT_HAVE_MATCHING_SHOP.to_string())
            }
            _ => None,
        }
    }

    fn create_app_command(command: &mut CreateApplicationCommand) -> &mut CreateApplicationCommand {
        command
            .name(Self::command_name())
            .description("Remove an item from a shop")
            .create_option(|option| {
                option
                    .name("shop_name")
                    .description("Shop remove an item")
                    .kind(ApplicationCommandOptionType::String)
                    .required(true)
            })
            .create_option(|option| {
                option
                    .name("item_name")
                    .description("Name of the item to remove")
                    .kind(ApplicationCommandOptionType::String)
                    .required(true)
            })
    }

    async fn process_arguments(
        command_interaction: ApplicationCommandInteraction,
    ) -> Result<Self::ApiParams, BotError> {
        let options = command_interaction.data.options;

        Ok(Self::ApiParams::new(
            option_to_string(options.get(0), "shop_name")?,
            option_to_string(options.get(1), "item_name")?,
        ))
    }

    fn build_response(_: &GeoffreyContext, resp: Self::ApiResp, args: Self::ApiParams) -> String {
        MessageBuilder::new()
            .push_bold_safe(&args.item_name)
            .push(" has been removed from ")
            .push_bold_safe(resp.name)
            .push_line("")
            .build()
    }
}
